@@ -39,8 +39,6 @@
 #include "simutil.h"
 #include "smsutil.h"
 
-#define MESSAGE_WAITING_INTERFACE "org.ofono.MessageWaiting"
-
 struct mailbox_state {
 	gboolean indication;
 	unsigned char message_count;
@@ -51,12 +49,15 @@ struct ofono_message_waiting {
 	unsigned char efmwis_length;
 	unsigned char efmbdn_length;
 	unsigned char efmbdn_record_id[5];
+	unsigned int efmbdn_watch;
 	unsigned char ef_cphs_mwis_length;
 	unsigned char ef_cphs_mbdn_length;
+	unsigned int ef_cphs_mbdn_watch;
 	gboolean mbdn_not_provided;
 	gboolean cphs_mbdn_not_provided;
 	struct ofono_phone_number mailbox_number[5];
 	struct ofono_sim *sim;
+	struct ofono_sim_context *sim_context;
 	struct ofono_atom *atom;
 };
 
@@ -121,7 +122,7 @@ static DBusMessage *mw_get_properties(DBusConnection *conn,
 	const char *number;
 
 	reply = dbus_message_new_method_return(msg);
-	if (!reply)
+	if (reply == NULL)
 		return NULL;
 
 	dbus_message_iter_init_append(reply, &iter);
@@ -204,7 +205,7 @@ static DBusMessage *set_cphs_mbdn(struct ofono_message_waiting *mw,
 	sim_adn_build(efmbdn, req->mw->ef_cphs_mbdn_length,
 			&req->number, NULL);
 
-	if (ofono_sim_write(mw->sim, SIM_EF_CPHS_MBDN_FILEID,
+	if (ofono_sim_write(mw->sim_context, SIM_EF_CPHS_MBDN_FILEID,
 			sync ? cphs_mbdn_sync_cb : mbdn_set_cb,
 			OFONO_SIM_FILE_STRUCTURE_FIXED,
 			mw_mailbox_to_cphs_record[mailbox],
@@ -252,13 +253,15 @@ static void mbdn_set_cb(int ok, void *data)
 		number = phone_number_to_string(old);
 
 		ofono_dbus_signal_property_changed(conn, path,
-						MESSAGE_WAITING_INTERFACE,
+						OFONO_MESSAGE_WAITING_INTERFACE,
 						property, DBUS_TYPE_STRING,
 						&number);
 	}
 
-	/* Make a single attempt at keeping the CPHS version of the file
-	 * in sync.  */
+	/*
+	 * Make a single attempt at keeping the CPHS version of the file
+	 * in sync.
+	 */
 	if (req->cphs == FALSE)
 		set_cphs_mbdn(req->mw, TRUE, req->mailbox,
 				phone_number_to_string(&req->number), NULL);
@@ -276,7 +279,7 @@ static DBusMessage *set_mbdn(struct ofono_message_waiting *mw, int mailbox,
 	struct mbdn_set_request *req;
 	unsigned char efmbdn[255];
 
-	/* 
+	/*
 	 * If we have no 3GPP EFmbdn on the card, maybe the
 	 * CPHS version is available
 	 */
@@ -300,10 +303,10 @@ static DBusMessage *set_mbdn(struct ofono_message_waiting *mw, int mailbox,
 
 	sim_adn_build(efmbdn, req->mw->efmbdn_length, &req->number, NULL);
 
-	if (ofono_sim_write(req->mw->sim, SIM_EFMBDN_FILEID, mbdn_set_cb,
-			OFONO_SIM_FILE_STRUCTURE_FIXED,
-			req->mw->efmbdn_record_id[mailbox],
-			efmbdn, req->mw->efmbdn_length, req) == -1) {
+	if (ofono_sim_write(req->mw->sim_context, SIM_EFMBDN_FILEID,
+				mbdn_set_cb, OFONO_SIM_FILE_STRUCTURE_FIXED,
+				req->mw->efmbdn_record_id[mailbox],
+				efmbdn, req->mw->efmbdn_length, req) == -1) {
 		g_free(req);
 
 		if (msg)
@@ -395,16 +398,16 @@ static void update_indicator_and_emit(struct ofono_message_waiting *mw,
 	indication = info->indication;
 	count = info->message_count;
 
-	if (!mw_message_waiting_property_name[mailbox])
+	if (mw_message_waiting_property_name[mailbox] == NULL)
 		return;
 
 	ofono_dbus_signal_property_changed(conn, path,
-				MESSAGE_WAITING_INTERFACE,
+				OFONO_MESSAGE_WAITING_INTERFACE,
 				mw_message_waiting_property_name[mailbox],
 				DBUS_TYPE_BOOLEAN, &indication);
 
 	ofono_dbus_signal_property_changed(conn, path,
-				MESSAGE_WAITING_INTERFACE,
+				OFONO_MESSAGE_WAITING_INTERFACE,
 				mw_message_count_property_name[mailbox],
 				DBUS_TYPE_BYTE, &count);
 }
@@ -418,7 +421,7 @@ static void mw_cphs_mwis_read_cb(int ok, int total_length, int record,
 	unsigned char indication;
 
 	if (!ok || total_length < 1) {
-		ofono_debug("No CPHS MWIS on SIM");
+		DBG("No CPHS MWIS on SIM");
 		mw->ef_cphs_mwis_length = 0;
 		return;
 	}
@@ -460,7 +463,7 @@ static void mw_mwis_read_cb(int ok, int total_length, int record,
 
 	if (!ok || record_length < 5) {
 		ofono_error("Unable to read waiting messages numbers "
-			"from SIM");
+				"from SIM");
 
 		mw->efmwis_length = 0;
 
@@ -526,7 +529,7 @@ static void mw_cphs_mbdn_read_cb(int ok, int total_length, int record,
 		value = phone_number_to_string(&mw->mailbox_number[i]);
 
 		ofono_dbus_signal_property_changed(conn, path,
-				MESSAGE_WAITING_INTERFACE,
+				OFONO_MESSAGE_WAITING_INTERFACE,
 				mw_mailbox_property_name[i],
 				DBUS_TYPE_STRING, &value);
 	}
@@ -543,7 +546,7 @@ static void mw_mbdn_read_cb(int ok, int total_length, int record,
 
 	if (!ok || record_length < 14 || total_length < record_length) {
 		ofono_error("Unable to read mailbox dialling numbers "
-			"from SIM");
+				"from SIM");
 
 		mw->efmbdn_length = 0;
 		mw->mbdn_not_provided = TRUE;
@@ -567,7 +570,7 @@ static void mw_mbdn_read_cb(int ok, int total_length, int record,
 		value = phone_number_to_string(&mw->mailbox_number[i]);
 
 		ofono_dbus_signal_property_changed(conn, path,
-				MESSAGE_WAITING_INTERFACE,
+				OFONO_MESSAGE_WAITING_INTERFACE,
 				mw_mailbox_property_name[i],
 				DBUS_TYPE_STRING, &value);
 	}
@@ -575,6 +578,40 @@ static void mw_mbdn_read_cb(int ok, int total_length, int record,
 	mw->efmbdn_length = record_length;
 }
 
+static void mw_mbdn_changed(int id, void *userdata)
+{
+	struct ofono_message_waiting *mw = userdata;
+	int err;
+
+	mw->efmbdn_length = 0;
+	mw->mbdn_not_provided = FALSE;
+
+	err = ofono_sim_read(mw->sim_context, SIM_EFMBDN_FILEID,
+				OFONO_SIM_FILE_STRUCTURE_FIXED,
+				mw_mbdn_read_cb, mw);
+	if (err != 0)
+		ofono_error("Unable to read EF-MBDN from SIM");
+}
+
+static void mw_cphs_mbdn_changed(int id, void *userdata)
+{
+	struct ofono_message_waiting *mw = userdata;
+
+	mw->ef_cphs_mbdn_length = 0;
+	mw->cphs_mbdn_not_provided = FALSE;
+
+	ofono_sim_read(mw->sim_context, SIM_EF_CPHS_MBDN_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_FIXED,
+			mw_cphs_mbdn_read_cb, mw);
+}
+
+const struct ofono_phone_number *__ofono_message_waiting_get_mbdn(
+					struct ofono_message_waiting *mw,
+					unsigned int index)
+{
+	return &mw->mailbox_number[index];
+}
+
 static void mw_mbi_read_cb(int ok, int total_length, int record,
 				const unsigned char *data,
 				int record_length, void *userdata)
@@ -585,7 +622,7 @@ static void mw_mbi_read_cb(int ok, int total_length, int record,
 
 	if (!ok || record_length < 4) {
 		ofono_error("Unable to read mailbox identifies "
-			"from SIM");
+				"from SIM");
 
 		mw->efmbdn_length = 0;
 		mw->mbdn_not_provided = TRUE;
@@ -600,9 +637,12 @@ static void mw_mbi_read_cb(int ok, int total_length, int record,
 	for (i = 0; i < 5 && i < record_length; i++)
 		mw->efmbdn_record_id[i] = data[i];
 
-	err = ofono_sim_read(mw->sim, SIM_EFMBDN_FILEID,
+	err = ofono_sim_read(mw->sim_context, SIM_EFMBDN_FILEID,
 				OFONO_SIM_FILE_STRUCTURE_FIXED,
 				mw_mbdn_read_cb, mw);
+	mw->efmbdn_watch = ofono_sim_add_file_watch(mw->sim_context,
+						SIM_EFMBDN_FILEID,
+						mw_mbdn_changed, mw, NULL);
 
 	if (err != 0)
 		ofono_error("Unable to read EF-MBDN from SIM");
@@ -614,10 +654,15 @@ out:
 	 */
 	st = ofono_sim_get_cphs_service_table(mw->sim);
 
-	if (st && bit_field(st[0], 4, 2) == 3)
-		ofono_sim_read(mw->sim, SIM_EF_CPHS_MBDN_FILEID,
+	if (st && bit_field(st[0], 4, 2) == 3) {
+		ofono_sim_read(mw->sim_context, SIM_EF_CPHS_MBDN_FILEID,
 				OFONO_SIM_FILE_STRUCTURE_FIXED,
 				mw_cphs_mbdn_read_cb, mw);
+		mw->ef_cphs_mbdn_watch = ofono_sim_add_file_watch(
+						mw->sim_context,
+						SIM_EF_CPHS_MBDN_FILEID,
+						mw_cphs_mbdn_changed, mw, NULL);
+	}
 }
 
 static void mw_mwis_write_cb(int ok, void *userdata)
@@ -654,7 +699,7 @@ static void mw_set_indicator(struct ofono_message_waiting *mw, int profile,
 
 		if (mw_message_waiting_property_name[type])
 			ofono_dbus_signal_property_changed(conn, path,
-					MESSAGE_WAITING_INTERFACE,
+					OFONO_MESSAGE_WAITING_INTERFACE,
 					mw_message_waiting_property_name[type],
 					DBUS_TYPE_BOOLEAN, &indication);
 	}
@@ -666,7 +711,7 @@ static void mw_set_indicator(struct ofono_message_waiting *mw, int profile,
 
 		if (mw_message_waiting_property_name[type])
 			ofono_dbus_signal_property_changed(conn, path,
-					MESSAGE_WAITING_INTERFACE,
+					OFONO_MESSAGE_WAITING_INTERFACE,
 					mw_message_count_property_name[type],
 					DBUS_TYPE_BYTE, &messages);
 	}
@@ -689,7 +734,8 @@ static void mw_set_indicator(struct ofono_message_waiting *mw, int profile,
 		if (mw->messages[i].indication)
 			efmwis[0] |= 1 << i;
 
-	if (ofono_sim_write(mw->sim, SIM_EFMWIS_FILEID, mw_mwis_write_cb,
+	if (ofono_sim_write(mw->sim_context, SIM_EFMWIS_FILEID,
+				mw_mwis_write_cb,
 				OFONO_SIM_FILE_STRUCTURE_FIXED, 1,
 				efmwis, mw->efmwis_length, mw) != 0) {
 		ofono_error("Queuing a EF-MWI write to SIM failed");
@@ -707,7 +753,8 @@ try_cphs:
 		efmwis[1] = mw->messages[1].indication ? 0xa : 0x5 |
 			mw->messages[3].indication ? 0xa0 : 0x50;
 
-	if (ofono_sim_write(mw->sim, SIM_EF_CPHS_MWIS_FILEID, mw_mwis_write_cb,
+	if (ofono_sim_write(mw->sim_context, SIM_EF_CPHS_MWIS_FILEID,
+				mw_mwis_write_cb,
 				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT, 0,
 				efmwis, mw->ef_cphs_mwis_length, mw) != 0)
 		ofono_error("Queuing a EF-MWIS write to SIM failed (CPHS)");
@@ -729,10 +776,12 @@ static void handle_special_sms_iei(struct ofono_message_waiting *mw,
 		if (type == (SMS_MWI_TYPE_OTHER | 4))
 			type = SMS_MWI_TYPE_VIDEO;
 		else
-			/* 23.040 9.2.3.24.2: "Terminals should be capable of
+			/*
+			 * 23.040 9.2.3.24.2: "Terminals should be capable of
 			 * receiving any values in octet 1, even including
 			 * those marked as Reserved."  Treat Reserved as
-			 * "Other".  */
+			 * "Other".
+			 */
 			type = SMS_MWI_TYPE_OTHER;
 	}
 
@@ -821,7 +870,8 @@ void __ofono_message_waiting_mwi(struct ofono_message_waiting *mw,
 	if (out_discard)
 		*out_discard = FALSE;
 
-	/* Check MWI types in the order from highest priority to lowest
+	/*
+	 * Check MWI types in the order from highest priority to lowest
 	 * because they must override one another.
 	 */
 
@@ -882,7 +932,8 @@ void __ofono_message_waiting_mwi(struct ofono_message_waiting *mw,
 		}
 
 		if (iei_found) {
-			/* 23.040 9.2.3.24.2 says "In the event of a
+			/*
+			 * 23.040 9.2.3.24.2 says "In the event of a
 			 * conflict between this setting and the setting
 			 * of the Data Coding Scheme (see 3GPP TS 23.038 [9])
 			 * then the message shall be stored if either the DCS
@@ -914,49 +965,118 @@ static void message_waiting_unregister(struct ofono_atom *atom)
 	DBusConnection *conn = ofono_dbus_get_connection();
 	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
 	const char *path = __ofono_atom_get_path(atom);
+	struct ofono_message_waiting *mw = __ofono_atom_get_data(atom);
+
+	if (mw->sim_context) {
+		ofono_sim_context_free(mw->sim_context);
+		mw->sim_context = NULL;
+	}
+
+	mw->sim = NULL;
 
 	g_dbus_unregister_interface(conn, path,
-					MESSAGE_WAITING_INTERFACE);
-	ofono_modem_remove_interface(modem, MESSAGE_WAITING_INTERFACE);
+					OFONO_MESSAGE_WAITING_INTERFACE);
+	ofono_modem_remove_interface(modem, OFONO_MESSAGE_WAITING_INTERFACE);
+}
+
+static void mw_mwis_changed(int id, void *userdata)
+{
+	struct ofono_message_waiting *mw = userdata;
+
+	mw->efmwis_length = 0;
+
+	ofono_sim_read(mw->sim_context, SIM_EFMWIS_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_FIXED,
+			mw_mwis_read_cb, mw);
+}
+
+static void mw_cphs_mwis_changed(int id, void *userdata)
+{
+	struct ofono_message_waiting *mw = userdata;
+
+	mw->ef_cphs_mwis_length = 0;
+
+	ofono_sim_read(mw->sim_context, SIM_EF_CPHS_MWIS_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
+			mw_cphs_mwis_read_cb, mw);
+}
+
+static void mw_mbi_changed(int id, void *userdata)
+{
+	struct ofono_message_waiting *mw = userdata;
+
+	mw->efmbdn_length = 0;
+	mw->mbdn_not_provided = FALSE;
+
+	mw->ef_cphs_mbdn_length = 0;
+	mw->cphs_mbdn_not_provided = FALSE;
+
+	ofono_sim_remove_file_watch(mw->sim_context, mw->efmbdn_watch);
+	ofono_sim_remove_file_watch(mw->sim_context, mw->ef_cphs_mbdn_watch);
+
+	ofono_sim_read(mw->sim_context, SIM_EFMBI_FILEID,
+			OFONO_SIM_FILE_STRUCTURE_FIXED,
+			mw_mbi_read_cb, mw);
 }
 
 void ofono_message_waiting_register(struct ofono_message_waiting *mw)
 {
-	DBusConnection *conn = ofono_dbus_get_connection();
-	const char *path = __ofono_atom_get_path(mw->atom);
-	struct ofono_modem *modem = __ofono_atom_get_modem(mw->atom);
+	DBusConnection *conn;
+	const char *path;
+	struct ofono_modem *modem;
 	struct ofono_atom *sim_atom;
 
+	if (mw == NULL)
+		return;
+
+	conn = ofono_dbus_get_connection();
+	modem = __ofono_atom_get_modem(mw->atom);
+	path = __ofono_atom_get_path(mw->atom);
+
 	if (!g_dbus_register_interface(conn, path,
-					MESSAGE_WAITING_INTERFACE,
+					OFONO_MESSAGE_WAITING_INTERFACE,
 					message_waiting_methods,
 					message_waiting_signals,
 					NULL, mw, NULL)) {
 		ofono_error("Could not create %s interface",
-				MESSAGE_WAITING_INTERFACE);
+				OFONO_MESSAGE_WAITING_INTERFACE);
 		return;
 	}
 
-	ofono_modem_add_interface(modem, MESSAGE_WAITING_INTERFACE);
+	ofono_modem_add_interface(modem, OFONO_MESSAGE_WAITING_INTERFACE);
 
 	sim_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SIM);
 
 	if (sim_atom) {
 		/* Assume that if sim atom exists, it is ready */
 		mw->sim = __ofono_atom_get_data(sim_atom);
+		mw->sim_context = ofono_sim_context_create(mw->sim);
 
 		/* Loads MWI states and MBDN from SIM */
-		ofono_sim_read(mw->sim, SIM_EFMWIS_FILEID,
+		ofono_sim_read(mw->sim_context, SIM_EFMWIS_FILEID,
 				OFONO_SIM_FILE_STRUCTURE_FIXED,
 				mw_mwis_read_cb, mw);
-		ofono_sim_read(mw->sim, SIM_EFMBI_FILEID,
+		ofono_sim_read(mw->sim_context, SIM_EFMBI_FILEID,
 				OFONO_SIM_FILE_STRUCTURE_FIXED,
 				mw_mbi_read_cb, mw);
 
 		/* Also read CPHS MWIS field */
-		ofono_sim_read(mw->sim, SIM_EF_CPHS_MWIS_FILEID,
+		ofono_sim_read(mw->sim_context, SIM_EF_CPHS_MWIS_FILEID,
 				OFONO_SIM_FILE_STRUCTURE_TRANSPARENT,
 				mw_cphs_mwis_read_cb, mw);
+
+		/*
+		 * The operator could send us SMS mwi updates, but let's be
+		 * extra careful and track the file contents too.
+		 */
+		ofono_sim_add_file_watch(mw->sim_context, SIM_EFMWIS_FILEID,
+						mw_mwis_changed, mw, NULL);
+		ofono_sim_add_file_watch(mw->sim_context,
+						SIM_EF_CPHS_MWIS_FILEID,
+						mw_cphs_mwis_changed, mw, NULL);
+
+		ofono_sim_add_file_watch(mw->sim_context, SIM_EFMBI_FILEID,
+						mw_mbi_changed, mw, NULL);
 	}
 
 	__ofono_atom_register(mw->atom, message_waiting_unregister);
